package info.textgrid.middleware;

import static org.junit.Assert.assertTrue;
import java.io.IOException;
import java.net.URL;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import org.apache.cxf.helpers.IOUtils;
import org.apache.http.HttpStatus;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;
import info.textgrid.utils.httpclient.TGHttpClient;
import info.textgrid.utils.httpclient.TGHttpResponse;

/**
 * <p>
 * Some online tests for the TextGrid OAIMPH service.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 */
@Ignore
public class OaiPmhTextgridOnlineTests {

  private static final String OAI_DC_PREFIX = "oai_dc";
  private static final String OAI_IDIOMMETS_PREFIX = "oai_idiom_mets";
  private static final String NO_SET = null;
  private static final String ERROR = " >>> ERROR";
  private static final String OK = " >>> OKIDOKI";
  private static final String TESTING = "\n >>> TESTING ";
  private static final String NO_THREAD_NAME = "";
  private static final String NOT_TESTED = " >>> NOT_TESTED";

  private static final String VERB_LIST_IDENTIFIERS = "ListIdentifiers";
  private static final String VERB_LIST_RECORDS = "ListRecords";
  private static final String VERB_GET_RECORD = "GetRecord";
  private static final String VERB_IDENTIFY = "Identify";
  private static final String VERB_LIST_METADATA_FORMATS = "ListMetadataFormats";
  private static final String VERB_LIST_SETS = "ListSets";

  // The OAIPMH host to be tested.
  // private static String host = "http://dev.textgridlab.org/1.0/tgoaipmh/";
  private static String host = "http://textgridlab.org/1.0/tgoaipmh/";

  // NOTE Test with "mvn jetty:run" in oaipmh-webapp module folder and SSH tunneling "ssh -L
  // 9302:localhost:9302 -l username textgrid-esx2.gwdg.de", please use original OAIPMH config file
  // from /etc/dhrep/oaipmh/oaipmh.textgrid.properties and copy it to
  // /etc/dhrep/oaipmh/oaipmh.config on your local machine!
  // private static String host = "http://localhost:8095/oaipmh/";
  // NOTE Use "mvn tomcat:run" here!
  // private static String host = "http://localhost:8080/oaipmh-webapp/";

  // Time to wait between the single queries using resumption tokens in milliseconds.
  private static final Long TIME = 50l;

  // **
  // PREPARATIONS
  // **

  /**
   * @throws Exception
   */
  @BeforeClass
  public static void setUpBeforeClass() throws Exception {

  }

  /**
   * @throws Exception
   */
  @AfterClass
  public static void tearDownAfterClass() throws Exception {}

  /**
   * @throws Exception
   */
  @Before
  public void setUp() throws Exception {}

  /**
   * @throws Exception
   */
  @After
  public void tearDown() throws Exception {}

  // **
  // TESTS
  // **

  /**
   * @throws IOException
   */
  @Test
  public void testGetVersion() throws IOException {

    System.out.println(TESTING + "testGetVersion()");

    String shouldStartWith = "oaipmh-core";

    TGHttpResponse httpResponse = getHttpResponse(host + "oai/version", NO_THREAD_NAME);
    int status = httpResponse.getStatusCode();

    String response = IOUtils.readStringFromStream(httpResponse.getInputstream());

    if (status != HttpStatus.SC_OK || !response.startsWith(shouldStartWith)) {
      System.err.println("\tstatus: " + status);
      System.err.println(ERROR + ": response should start with '" + shouldStartWith + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testRootUrl() throws IOException {

    System.out.println(TESTING + "testRootUrl()");

    String shouldContain = "www.textgridrep.de";

    TGHttpResponse httpResponse = getHttpResponse(host + "oai", NO_THREAD_NAME);
    int status = httpResponse.getStatusCode();

    String response = IOUtils.readStringFromStream(httpResponse.getInputstream());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err.println(ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testIdentify() throws IOException {

    System.out.println(TESTING + "testIdentify()");

    String shouldContain = "www.textgridrep.de";

    TGHttpResponse httpResponse =
        getHttpResponse(host + "oai?verb=" + VERB_IDENTIFY, NO_THREAD_NAME);
    int status = httpResponse.getStatusCode();

    String response = IOUtils.readStringFromStream(httpResponse.getInputstream());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err.println(ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListMetadataFormats() throws IOException {

    System.out.println(TESTING + "testListMetadataFormats()");

    String shouldContainOaidc = "oai_dc";
    String shouldContainOaiidiommets = "oai_idiom_mets";

    TGHttpResponse httpResponse =
        getHttpResponse(host + "oai?verb=" + VERB_LIST_METADATA_FORMATS, NO_THREAD_NAME);
    int status = httpResponse.getStatusCode();

    String response = IOUtils.readStringFromStream(httpResponse.getInputstream());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContainOaidc)
        || !response.contains(shouldContainOaiidiommets)) {
      System.err.println("\tstatus: " + status);
      System.err.println(ERROR + ": response should contain '" + shouldContainOaidc + "' and '"
          + shouldContainOaiidiommets + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListSets() throws IOException {

    System.out.println(TESTING + "testListSets()");

    // Testing for set: Digital Bibliothek.
    String shouldContainTEXTGRIDLABORG = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";
    String shouldContainDEFTEXTGRIDLABORG = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";

    TGHttpResponse httpResponse =
        getHttpResponse(host + "oai?verb=" + VERB_LIST_SETS, NO_THREAD_NAME);
    int status = httpResponse.getStatusCode();

    String response = IOUtils.readStringFromStream(httpResponse.getInputstream());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContainTEXTGRIDLABORG)
        && !response.contains(shouldContainDEFTEXTGRIDLABORG)) {
      System.err.println("\tstatus: " + status);
      System.err.println(
          ERROR + ": response should contain (textgridlab.org) '" + shouldContainTEXTGRIDLABORG
              + "' or (dev.textgridlab.org) '" + shouldContainDEFTEXTGRIDLABORG + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDC() throws IOException {

    System.out.println(TESTING + "testGetRecordOAIDC()");

    String identifier = "textgrid:mq05.0";
    String shouldContain = "tvitt@textgrid.de";

    String url =
        host + "oai?verb=" + VERB_GET_RECORD + "&identifier=" + identifier + "&metadataPrefix="
            + OAI_DC_PREFIX;

    if (!host.contains("dev.textgridlab.org")) {
      TGHttpResponse httpResponse = getHttpResponse(url, NO_THREAD_NAME);
      int status = httpResponse.getStatusCode();
      String response = IOUtils.readStringFromStream(httpResponse.getInputstream());
      System.out.println("\tresponse: " + response);

      if (!response.contains(shouldContain)) {
        assertTrue(false);
      }

      System.out.println(OK);
    } else {
      System.out.println(NOT_TESTED);
    }
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS() throws IOException {

    System.out.println(TESTING + "testGetRecordIDIOMMETS()");

    String identifier = "textgrid:254w6";
    String shouldContain = "Tikal";

    String url =
        host + "oai?verb=" + VERB_GET_RECORD + "&identifier=" + identifier + "&metadataPrefix="
            + OAI_IDIOMMETS_PREFIX;

    if (!host.contains("dev.textgridlab.org")) {
      TGHttpResponse httpResponse = getHttpResponse(url, NO_THREAD_NAME);
      int status = httpResponse.getStatusCode();
      String response = IOUtils.readStringFromStream(httpResponse.getInputstream());
      System.out.println("\tresponse: " + response);

      if (!response.contains(shouldContain)) {
        assertTrue(false);
      }

      System.out.println(OK);
    } else {
      System.err.println(NOT_TESTED);
    }
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCSinglePage() throws IOException {

    System.out.println(TESTING + "testListRecordsOAIDCSinglePage()");

    // IIIF
    String set = "project:TGPR-2e532567-c3e3-36f5-3ecc-520cea7904cc";

    testList(VERB_LIST_RECORDS, set, OAI_DC_PREFIX, 0, 100, NO_THREAD_NAME);

    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCMorePages() throws IOException {

    System.out.println(TESTING + "testListRecordsOAIDCMorePages()");

    // Testing set: Digitale Bibliothek (TODO Find another project with less records for testing!).
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";
    if (host.contains("dev.textgridlab.org")) {
      // Use fu's test project, if test instance used in host name.
      // Testing set: fu's shakespeare.
      set = "project:TGPR-962c949c-56a3-7134-9265-582f29a0992d";
    }

    testList(VERB_LIST_RECORDS, set, OAI_DC_PREFIX, 0, 100, NO_THREAD_NAME);

    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsIDIOMMETSMorePages() throws IOException {

    System.out.println(TESTING + "testListRecordsIDIOMMETSMorePages()");

    // Do not test on dev.textgridlab.org.
    if (!host.contains("dev.textgridlab.org")) {
      testList(VERB_LIST_RECORDS, NO_SET, OAI_IDIOMMETS_PREFIX, 100, 30, NO_THREAD_NAME);
    }

    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifierOAIDCSinglePage() throws IOException {

    System.out.println(TESTING + "testListIdentifierOAIDCSinglePage()");

    // IIIF
    String set = "project:TGPR-2e532567-c3e3-36f5-3ecc-520cea7904cc";

    testList(VERB_LIST_IDENTIFIERS, set, OAI_DC_PREFIX, 0, 100, NO_THREAD_NAME);

    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersOAIDCMorePages() throws IOException {

    System.out.println(TESTING + "testListIdentifiersOAIDCMorePages()");

    // Testing set: Digitale Bibliothek (TODO Find another project with less records for testing!).
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";
    if (host.contains("dev.textgridlab.org")) {
      // Use fu's test project, if test instance used in host name.
      // Testing set: fu's shakespeare.
      set = "project:TGPR-962c949c-56a3-7134-9265-582f29a0992d";
    }

    testList(VERB_LIST_IDENTIFIERS, set, OAI_DC_PREFIX, 0, 100, NO_THREAD_NAME);

    System.out.println(OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersIDIOMMETSMorePages() throws IOException {

    System.out.println(TESTING + "testListIdentifiersIDIOMMETSMorePages()");

    testList(VERB_LIST_IDENTIFIERS, NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, NO_THREAD_NAME);

    System.out.println(OK);
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListIdentifiersConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out.println(TESTING + "testListIdentifiersConcurrentlyIDIOMMETSMorePages()");

    // Do not test on dev.textgridlab.org.
    if (!host.contains("dev.textgridlab.org")) {
      ExecutorService executor = Executors.newFixedThreadPool(3);

      Future<Boolean> f1 =
          executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
              NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[C1]"));
      Future<Boolean> f2 =
          executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
              NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[C2]"));
      Future<Boolean> f3 =
          executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
              NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[C2]"));

      executor.shutdown();

      System.out.println(OK + ": [C1]=" + f1.get() + ", [C2]=" + f2.get() + ", [C3]=" + f3.get());
    } else {
      System.out.println(NOT_TESTED);
    }
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListRecordsConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out.println(TESTING + "testListRecordsConcurrentlyIDIOMMETSMorePages()");

    // Do not test on dev.textgridlab.org.
    if (!host.contains("dev.textgridlab.org")) {
      ExecutorService executor = Executors.newFixedThreadPool(3);

      Future<Boolean> f1 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
          NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[D1]"));
      Future<Boolean> f2 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
          NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[D2]"));
      Future<Boolean> f3 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
          NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[D3]"));

      executor.shutdown();

      System.out.println(OK + ": [D1]=" + f1.get() + ", [D2]=" + f2.get() + ", [D3]=" + f3.get());

    } else {
      System.out.println(NOT_TESTED);
    }
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiers()
      throws InterruptedException, ExecutionException {

    System.out.println(TESTING + "testRestokConcurrentlyListIdentifiers()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_DC_PREFIX, 200, 100, "[A1]"));
    Future<Boolean> f2 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_DC_PREFIX, 201, 100, "[A2]"));
    Future<Boolean> f3 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_DC_PREFIX, 210, 100, "[A3]"));

    executor.shutdown();

    System.out.println(OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get() + ", [A3]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out.println(TESTING + "testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_DC_PREFIX, 201, 100, "[DC1]"));
    Future<Boolean> f2 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_DC_PREFIX, 199, 100, "[DC2]"));
    Future<Boolean> f3 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_IDENTIFIERS,
        NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[IDIOM3]"));

    executor.shutdown();

    System.out.println(OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get() + ", [A3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecords() throws InterruptedException, ExecutionException {

    System.out.println(TESTING + "testRestokConcurrentlyListRecords()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_DC_PREFIX, 33, 100, "[B1]"));
    Future<Boolean> f2 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_DC_PREFIX, 33, 100, "[B2]"));
    Future<Boolean> f3 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_DC_PREFIX, 33, 100, "[B3]"));

    executor.shutdown();

    System.out.println(OK + ": [B1]=" + f1.get() + ", [B2]=" + f2.get() + ", [B3]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecordsDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out.println(TESTING + "testRestokConcurrentlyListRecordsAndIDIOMMets()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_DC_PREFIX, 66, 100, "[DC1]"));
    Future<Boolean> f2 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_DC_PREFIX, 77, 100, "[DC2]"));
    Future<Boolean> f3 = executor.submit(new OaiPmhResumptionTokenTestThread(VERB_LIST_RECORDS,
        NO_SET, OAI_IDIOMMETS_PREFIX, 3, 30, "[IDIOM1]"));

    executor.shutdown();

    System.out
        .println(OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get() + ", [IOIOM1]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testConcurrentlyOAIDCGetRecord() throws InterruptedException, ExecutionException {

    System.out.println(TESTING + "testConcurrentlyOAIDCGetRecord()");

    String id1 = "textgrid:mq05.0";
    String id2 = "textgrid:jgv6.0";
    String id3 = "textgrid:w7rz.0";
    String id4 = "textgrid:w36b.0";
    String id5 = "textgrid:mcdv.0";
    String id6 = "textgrid:v0qx.0";
    String id7 = "textgrid:xkck.0";
    String id8 = "textgrid:t3m1.0";
    String id9 = "textgrid:10rsq.0";
    String id10 = "textgrid:mq05.0";
    String id11 = "textgrid:jgv6.0";
    String id12 = "textgrid:w7rz.0";
    String id13 = "textgrid:w36b.0";
    String id14 = "textgrid:mcdv.0";
    String id15 = "textgrid:v0qx.0";
    String id16 = "textgrid:xkck.0";
    String id17 = "textgrid:t3m1.0";
    String id18 = "textgrid:10rsq.0";

    if (!host.contains("dev.textgridlab.org")) {

      ExecutorService executor = Executors.newFixedThreadPool(18);

      Future<Boolean> f1 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id1, OAI_DC_PREFIX, "[GR1]"));
      Future<Boolean> f2 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id2, OAI_DC_PREFIX, "[GR2]"));
      Future<Boolean> f3 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id3, OAI_DC_PREFIX, "[GR3]"));
      Future<Boolean> f4 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id4, OAI_DC_PREFIX, "[GR4]"));
      Future<Boolean> f5 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id5, OAI_DC_PREFIX, "[GR5]"));
      Future<Boolean> f6 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id6, OAI_DC_PREFIX, "[GR6]"));
      Future<Boolean> f7 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id7, OAI_DC_PREFIX, "[GR7]"));
      Future<Boolean> f8 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id8, OAI_DC_PREFIX, "[GR8]"));
      Future<Boolean> f9 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id9, OAI_DC_PREFIX, "[GR9]"));
      Future<Boolean> f10 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id10, OAI_DC_PREFIX, "[GR10]"));
      Future<Boolean> f11 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id11, OAI_DC_PREFIX, "[GR11]"));
      Future<Boolean> f12 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id12, OAI_DC_PREFIX, "[GR12]"));
      Future<Boolean> f13 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id13, OAI_DC_PREFIX, "[GR13]"));
      Future<Boolean> f14 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id14, OAI_DC_PREFIX, "[GR14]"));
      Future<Boolean> f15 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id15, OAI_DC_PREFIX, "[GR15]"));
      Future<Boolean> f16 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id16, OAI_DC_PREFIX, "[GR16]"));
      Future<Boolean> f17 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id17, OAI_DC_PREFIX, "[GR17]"));
      Future<Boolean> f18 = executor
          .submit(new OaiPmhGetRecordTestThread(VERB_GET_RECORD, id18, OAI_DC_PREFIX, "[GR18]"));

      executor.shutdown();

      System.out.println(OK + "[GR1]=" + f1.get() + ", [GR2]=" + f2.get() + ", [GR3]=" + f3.get()
          + ", [GR4]=" + f4.get() + ", [GR5]=" + f5.get() + ", [GR6]=" + f6.get() + ", [GR7]="
          + f7.get() + ", [GR8]=" + f8.get() + ", [GR9]=" + f9.get() + ", [GR10]=" + f10.get()
          + ", [GR11]=" + f11.get() + ", [GR12]=" + f12.get() + ", [GR13]=" + f13.get()
          + ", [GR14]=" + f14.get() + ", [GR15]=" + f15.get() + ", [GR16]=" + f16.get()
          + ", [GR17]=" + f17.get() + ", [GR18]=" + f18.get());
    } else {
      System.out.println(NOT_TESTED);
    }
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * @param theUrl
   * @param theThreadName
   * @return
   * @throws IOException
   */
  private static TGHttpResponse getHttpResponse(String theUrl, String theThreadName)
      throws IOException {

    try {
      Thread.sleep(TIME);
    } catch (InterruptedException e) {
      // Nothing interrupts here!
    }

    System.out.println("\t" + theThreadName + " request:  " + theUrl);

    URL url = new URL(theUrl);
    String host = url.getProtocol() + "://" + url.getHost() + ":" + url.getPort();
    String query = url.getPath() + "?" + url.getQuery();
    TGHttpClient client = new TGHttpClient(host);

    // Get response.
    TGHttpResponse response = client.get(query);

    // Check HTTP status.
    int status = response.getStatusCode();
    String reason = response.getReasonPhrase();

    if (status != HttpStatus.SC_OK) {
      System.err.println("\t" + theThreadName + " status:   " + status + " " + reason);
      System.err.println(ERROR);
      assertTrue(false);
    } else {
      System.out.println("\t" + theThreadName + " status:   " + status + " " + reason);
    }

    return response;
  }

  /**
   * @param theResponse
   * @param recordOrHeader
   * @param oldtok
   * @param recordsExpectedPerRequest
   * @param loopCount
   * @param theThreadName
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  private static String examineResumptionTokenTag(TGHttpResponse theResponse, String recordOrHeader,
      String oldtok, int recordsExpectedPerRequest, int loopCount, String theThreadName)
      throws IOException {

    String res = IOUtils.readStringFromStream(theResponse.getInputstream());
    // Test for OAIPMH errors.
    if (res.contains("<error code=\"badArgument\">")) {
      System.err.println(theThreadName + ERROR + " IN OAIPMH RESPONSE: " + res);
      assertTrue(false);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = res.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = res.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + theThreadName + " " + recordOrHeader + "s:  " + recordCount);

    // Check if token tag is existing.
    int tokStart = res.indexOf("<resumptionToken");
    int tokEnd = res.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\t" + theThreadName + " token:    no token");

      return "-1";
    }

    String restokTmp = res.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + " tokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + " token:    " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\t" + theThreadName + " tokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\t" + theThreadName + " size:     " + size + " / " + cursor);

    // If token is provided, test cursor and element count!
    if (!restok.isEmpty()) {
      synchronized (OaiPmhTextgridOnlineTests.class) {
        // Check <record> or <header> count, must be recordsExpectedPerRequest!
        if (recordCount != recordsExpectedPerRequest) {
          String message = ERROR + ": " + recordOrHeader + " count mismatch, must be "
              + recordsExpectedPerRequest + " if token is provided, but is " + recordCount;
          assertTrue(message, false);
        }
        if (size <= recordsExpectedPerRequest) {
          String message = ERROR + ": completeListSize count mismatch, must be > "
              + recordsExpectedPerRequest + " if token is provided, but is " + size;
          assertTrue(message, false);
        }
        if (cursor != recordsExpectedPerRequest * loopCount) {
          String message = ERROR + ": cursor must be " + (loopCount * recordsExpectedPerRequest)
              + " in loop " + loopCount + ", but is " + cursor;
          assertTrue(message, false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % recordsExpectedPerRequest.
      if (recordCount != size % recordsExpectedPerRequest) {
        System.err
            .println(theThreadName + ERROR + ": " + recordOrHeader + " count mismatch, should be "
                + size % recordsExpectedPerRequest + ", but is " + recordCount);
        assertTrue(false);
      }

      // No resumption token available in response.
      return "-1";
    }

    return restok;
  }

  /**
   * @param theMillis
   * @return
   */
  public static String getDurationInSecs(long theMillis) {

    int SECS_IN_MILLIS = 1000;
    int secs = SECS_IN_MILLIS;

    return theMillis / secs + " second" + ((theMillis / secs) != 1 ? "s" : "");
  }

  /**
   * @param theVerb
   * @param theSet
   * @param metadataPrefix
   * @param numberOfPagesToTest
   * @param recordsExpectedPerRequest
   * @param threadName
   * @throws IOException
   */
  protected static void testList(final String theVerb, final String theSet,
      final String theMetadataPrefix, final int numberOfPagesToTest,
      final int recordsExpectedPerRequest, final String theThreadName) throws IOException {

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String url = host + "oai?verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null) {
      url += "&set=" + theSet;
    }

    TGHttpResponse httpResponse = getHttpResponse(url, theThreadName);
    int status = httpResponse.getStatusCode();

    int loopCount = 1;
    long timeRunning = System.currentTimeMillis() - startTime;
    System.out
        .println("\t" + theThreadName + " time:     " + getDurationInSecs(timeRunning) + " (loop "
            + loopCount + (numberOfPagesToTest != 0 ? "/" + numberOfPagesToTest : "") + ")");

    String restok = examineResumptionTokenTag(httpResponse, testOccurance, "",
        recordsExpectedPerRequest, loopCount, theThreadName);

    while (status == HttpStatus.SC_OK && !restok.equals("-1")) {
      loopCount += 1;
      synchronized (theThreadName) {
        if (numberOfPagesToTest > 0 && loopCount > numberOfPagesToTest) {
          System.out
              .println("\t" + theThreadName + " TESTING ONLY " + numberOfPagesToTest + " pages!");
          break;
        }
        url = host + "oai?verb=" + theVerb + "&resumptionToken=" + restok;
        httpResponse = getHttpResponse(url, theThreadName);
        timeRunning = System.currentTimeMillis() - startTime;
        System.out
            .println("\t" + theThreadName + " time:     " + getDurationInSecs(timeRunning)
                + " (loop "
                + loopCount + (numberOfPagesToTest != 0 ? "/" + numberOfPagesToTest : "") + ")");
        restok = examineResumptionTokenTag(httpResponse, testOccurance, restok,
            recordsExpectedPerRequest, loopCount, theThreadName);

      }
    }

    if (loopCount < numberOfPagesToTest) {
      System.out.println(theThreadName + ERROR + ": " + "Must have done " + numberOfPagesToTest
          + " loops, but did only " + loopCount);
      assertTrue(false);
    }

    System.out.println("\t" + theThreadName + OK);
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhResumptionTokenTestThread implements Callable<Boolean> {

    protected String verb;
    protected String set;
    protected String metadataPrefix;
    protected int numberOfPagesToTest;
    protected int recordsExpectedPerRequest;
    protected String threadName;

    /**
     * @param theVerb
     * @param theSet
     * @param theMetadataPrefix
     * @param numberOfPagesToTest
     * @param recordsExpectedPerRequest
     */
    public OaiPmhResumptionTokenTestThread(String theVerb, String theSet, String theMetadataPrefix,
        int numberOfPagesToTest, int recordsExpectedPerRequest, String theThreadName) {
      this.verb = theVerb;
      this.set = theSet;
      this.metadataPrefix = theMetadataPrefix;
      this.numberOfPagesToTest = numberOfPagesToTest;
      this.recordsExpectedPerRequest = recordsExpectedPerRequest;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    public Boolean call() throws IOException {
      OaiPmhTextgridOnlineTests.testList(this.verb, this.set, this.metadataPrefix,
          this.numberOfPagesToTest, this.recordsExpectedPerRequest, this.threadName);
      return true;
    }

  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhGetRecordTestThread implements Callable<Boolean> {

    protected String verb;
    protected String identifier;
    protected String metadataPrefix;
    protected String threadName;

    /**
     * @param theVerb
     * @param theIdentifier
     * @param theMetadataPrefix
     * @param theThreadName
     */
    public OaiPmhGetRecordTestThread(String theVerb, String theIdentifier, String theMetadataPrefix,
        String theThreadName) {
      this.verb = theVerb;
      this.identifier = theIdentifier;
      this.metadataPrefix = theMetadataPrefix;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.concurrent.Callable#call()
     */
    public Boolean call() throws IOException {

      String url = host + "oai?verb=" + VERB_GET_RECORD + "&identifier=" + this.identifier
          + "&metadataPrefix=" + this.metadataPrefix;

      TGHttpResponse httpResponse = getHttpResponse(url, this.threadName);

      return true;
    }
  }

}
